/* Funcoes graficas basicas para operar video.
Assembler AT&T + DJGPP:
- se gravar em variavel de memoria passada pelos parametros
  por "memory" no 3 :;
- __volatile__ permite ao compilador reorganizar o codigo
*/



#include <go32.h>
#include <string.h>   // necessario para acessar sys/movedata.h
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <pc.h>

#include "defs.h"

#include "Pol2d.h"

#include "gr.h"



// calcula declividade da reta entre dois vertices
#define DECLIVIDADE( XI, YI, XF, YF ) ( (float) ( ( YF ) - ( YI ) ) / ( ( XF ) - ( XI ) ) )

// calcula incremento de x (inverso da declividade)
#define INCREMENTOX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* construtor - Aloca memoria, inicia modo grafico, programa palette,...
**********************************************************/
CGr::CGr ()
{
	// tenta alocar memoria para memoria virtual 0, se nao conseguir sai
	if ( ( m_pyVr [ 0 ] = new BYTE [ TAMVR ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

	// tenta alocar memoria para memoria virtual 1, se nao conseguir sai
	if ( ( m_pyVr [ 1 ] = new BYTE [ TAMVR ] ) == NULL ) {
   	delete m_pyVr [ 0 ];
      m_pyVr [ 0 ] == NULL;
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

	// inicia com a memoria virtual 0 ativa
	m_nVrAtiva = 0;
	mAtrVrCor ( 0 );

	// entra em modo grafico
	mAtrVdModo ( MCGA );

	// faz com que a proporcao entre X e Y dos pixels seja 1
	outportb ( 0x3C2, 0xE3 );

	// programa palette
	for ( INT ncG = MINCOR; ncG < ( MAXCOR + 1 ); ncG ++ )
		mAtrPalCor ( ncG, 0, 0, ncG );
}



/* destrutor - Volta ao modo texto, desaloca memoria
**********************************************************/
CGr::~CGr ()
{
	// entra em modo texto
	mAtrVdModo ( TEXTO );

	// desaloca
	delete m_pyVr [ 0 ];
   if ( m_pyVr [ 1 ] != NULL )
		delete m_pyVr [ 1 ];
}



/* Altera o modo de exibicao do video.
**********************************************************/
void CGr::mAtrVdModo ( UINT16 uModo )
{
	REGS
		stRegsE, stRegsS;      // registradores

	stRegsE.x.ax = uModo;
	int86 ( 0x10, &stRegsE, &stRegsS );
}



/* Atribui uma quantidade de RGB para uma cor (entrada) da palette
**********************************************************/
void CGr::mAtrPalCor ( BYTE yCor, BYTE yR, BYTE yG, BYTE yB )
{
	outportb ( 0x03C8, yCor );
	outportb ( 0x03C9, yR );
	outportb ( 0x03C9, yG );
	outportb ( 0x03C9, yB );
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
**********************************************************/
void CGr::mVdSincRetraco ()
{
	__asm__ __volatile__ ("

			 movw  $0x03DA, %%dx
		L1: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jnz   L1

		L2: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    L2
		"
		:
		:
		: "%dx", "%al"
	);
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
  durante a espera zera a pagina de video virtual nao ativa.
__asm__ __volatile__ ("

		movl  %0,      %%edi   edi com endereco inicial a ser zerado
		movl  $16000,  %%ecx   ecx recebe o numero de interacoes (16000)
		xorl  %%ebx,   %%ebx   ebx recebe 0 (valor a ser escrito)

		movw  $0x03DA, %%dx    parte do "WaintRetrace"

  A1: movl  %%ebx,  (%%edi)  escreve no endereco apontado por edi um zero de 4 bytes (conteudo de ebx)
		addl  $4,      %%edi   soma 4 bytes no edi, proximos 4 bytes a serem escritos
		subl  $1,      %%ecx   subtrai 1 de ecx, menos uma interacao
		cmpl  $0,      %%ecx   se nao existe mais memoria para zerar contador de interacoes (ecx) chegou a zero nao zera mais a memoria
		jz    A2               se ecx == 0 desvia para parte onde soh faz primeiro teste do "WaitRetrace"

		inb   %%dx,    %%al    primeiro teste do "WaitRetrace"
		andb  $0x8,    %%al
		jz    A1               se o teste nao teve sucesso volta

  A4: movl  %%ebx,  (%%edi)  mesma escrita em endereco que em A1, soh depois de passar primeiro teste do "WaitRetrace"
		addl  $4,      %%edi
		subl  $1,      %%ecx
		cmpl  $0,      %%ecx
		jz    A3               se terminou de zerar memoria desvia para parte sem escrita em memoria (soh teste)

		inb   %%dx,    %%al    segundo teste do "WaitRetrace"
		andb  $0x8,    %%al
		jz    A4               se falhou teste volta

		jmp   AFimAtr          se acertou teste termina escrita na memoria antes de sair

  A2: inb   %%dx,    %%al    primeiro teste do "WaitRetrace" sem zerar memoria
		andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
		jnz   A2

  A3: inb   %%dx,    %%al    segundo teste do "WaitRetrace" sem zerar memoria
		andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
		jz    A3

		jmp AFim               termina sem passar por termino de escrita em memoria (logo abaixo), escrita jah acabou

  AFimAtr:
		movl  %%ebx,  (%%edi)  termina de escrever a memoria (apohs ter passado pelos 2 testes de "WaitRetrace")
		addl  $4,      %%edi
		subl  $1,      %%ecx
		cmpl  $0,      %%ecx
		jnz   AFimAtr

  AFim:

**********************************************************/
void CGr::mVrSincRetraco ()
{
   __asm__ __volatile__ ("

			 movl  %0,      %%edi
			 movl  $16000,  %%ecx
			 xorl  %%ebx,   %%ebx

			 movw  $0x03DA, %%dx

	   A1: movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jz    A2

			 inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A1

	   A4: movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jz    A3

			 inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A4

			 jmp   AFimAtr

	   A2: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jnz   A2

	   A3: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A3

			 jmp   AFim

	   AFimAtr:
			 movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jnz   AFimAtr

	   AFim:
	   "
	   :
	   : "g" ( m_pyVr [ ! m_nVrAtiva ] )
	   : "%dx", "%al", "%edi", "%ecx", "%eax", "%ebx"
   );
}



/* Atribui toda a tela com uma soh cor. Sincroniza com retraco.
**********************************************************/
void CGr::mAtrVdCorSinc ( BYTE yCor )
{
	BYTE
		ayCores [ 4 ];
	ayCores [ 0 ] = ayCores [ 1 ] = ayCores [ 2 ] = ayCores [ 3 ] = yCor;

	mVdSincRetraco ();

	for ( INT ncTrans = 0; ncTrans < TAMVR; ncTrans += 4 )
		_dosmemputl ( (void*) ayCores, 1, 0xA0000 + ncTrans );
}



/* Marca um pixel na tela
**********************************************************/
void CGr::mAtrVdPixel ( INT nX, INT nY, BYTE yCor )
{
	dosmemput ( (void*) &yCor, 1, 0xA0000 + nX + ( nY * RESX ) );
}



/* Atribui toda tela virtual com uma soh cor.
**********************************************************/
void CGr::mAtrVrCor ( BYTE yCor )
{
	memset ( (void*) m_pyVr [ m_nVrAtiva ], yCor, TAMVR );
}



/* Marca um pixel na memoria virtual
**********************************************************/
void CGr::mAtrVrPixel ( INT nX, INT nY, BYTE yCor )
{
	memset ( (void*) m_pyVr [ m_nVrAtiva ] + nX + ( nY * RESX ), yCor, 1 );
}



/* Transfere a memoria virtual para a tela, com sincronizacao
 de retraco. Ativa tela virtual inativa e desativa a ativa.
**********************************************************/
void CGr::mAtrVdVrSinc ()
{
	mVrSincRetraco ();

	// move dados de 4 em 4 bytes
	_movedatal ( _my_ds (), (INT) m_pyVr [ m_nVrAtiva ], _dos_ds, 0xA0000, TAMVR / 4 );

	// altera memoria virtual de video ativa para a que estava inativa e limpa
	m_nVrAtiva = ! m_nVrAtiva;
}



/* 1 - Desenha um poligono na memoria virtual de video usando
o incremento de x para calcular posicao de x. Entrada eh vertice por
vertice. Para garantir desenho de poligino convexo os vertices
adjacentes devem ocupar posicoes contiguas nas matrizes de vertices
do objeto poligono. Ou seja, visualmente devem ser definidos
em ordem horaria ou anti-horaria. Exemplo:

								(x4, y4)
								*                -
					        / \                 loop
							 /   \                    |
				(x1, y1) *     \                   |
							 \		 * (x3, y3)         | vai calculando posicoes de x
					        \	/                   | entre retas horizontais e
								\ /                    V retas das bordas do poligono
								 *               -
								 (x2, y2)

Equacao das retas:
Y = aX + b, onde
            a eh declividade
            b eh deslocamento da reta no eixo y
Equacao usada nas interseccoes:
X = ( deltaY * incrementoX ) + Xvertice, onde
     		   incrementoX eh 1 / a
		   	deltaY eh a distancia entre o Y de um vertice da reta e Y do loop
            Xvertice eh o X do vertice usado no calculo de deltaY

Os vertices sao numerados de 1 a 4 correspondendo a sua ordem
na matriz de vertices do objeto poligono.
Sao chamados vertices superior ou inferior os vertices dos extremos
considerando y, no exemplo sao (x4, y4) e (x2, y2), respectivamente.
**********************************************************/
void CGr::mAtrVrPol2D ( INT nX1, INT nY1, INT nX2, INT nY2, INT nX3, INT nY3,
								 INT nX4, INT nY4, BYTE yCor )
{
   // recorta poligono, faz com que vertices permanecam dentro da tela
	// verifica se os X estao fora dos limites, se estao os corrige
   // testa x1
	if ( nX1 < MINX )
		nX1 = MINX;
	else
		if ( nX1 > MAXX )
			nX1 = MAXX;
   // testa x2
	if ( nX2 < MINX )
		nX2 = MINX;
	else
		if ( nX2 > MAXX )
			nX2 = MAXX;
   // testa x3
	if ( nX3 < MINX )
		nX3 = MINX;
	else
		if ( nX3 > MAXX )
			nX3 = MAXX;
   // testa x4
	if ( nX4 < MINX )
		nX4 = MINX;
	else
		if ( nX4 > MAXX )
			nX4 = MAXX;
	// verifica se os Y estao fora dos limites, se estao os corrige
   // testa y1
	if ( nY1 < MINY )
		nY1 = MINY;
	else
		if ( nY1 > MAXY )
			nY1 = MAXY;
   // testa y2
	if ( nY2 < MINY )
		nY2 = MINY;
	else
		if ( nY2 > MAXY )
			nY2 = MAXY;
   // testa y3
	if ( nY3 < MINY )
		nY3 = MINY;
	else
		if ( nY3 > MAXY )
			nY3 = MAXY;
   // testa y4
	if ( nY4 < MINY )
		nY4 = MINY;
	else
		if ( nY4 > MAXY )
			nY4 = MAXY;

	// define maior e menor Y (vertices extremos em y).
	INT
		nYSup = nY1,
		nYInf = nY2;
   // se o valor do y superior for mais abaixo, na tela, entao muda
   if ( nYSup > nY1 )
      nYSup = nY1;
   if ( nYSup > nY2 )
      nYSup = nY2;
   if ( nYSup > nY3 )
      nYSup = nY3;
   if ( nYSup > nY4 )
      nYSup = nY4;
   // se o valor do y inferior for mais acima, na tela, entao muda
   if ( nYInf < nY1 )
      nYInf = nY1;
   if ( nYInf < nY2 )
      nYInf = nY2;
   if ( nYInf < nY3 )
      nYInf = nY3;
   if ( nYInf < nY4 )
      nYInf = nY4;

   // define se eh possivel calcular incrementos de x e se possivel calcula
   float
   	fIncX12 = 0,
   	fIncX14 = 0,
   	fIncX23 = 0,
   	fIncX34 = 0;
   BOOL
   	bIncValido12 = TRUE,
   	bIncValido14 = TRUE,
   	bIncValido23 = TRUE,
   	bIncValido34 = TRUE;
   // sao calculados valores apenas entre pontos que nao estao em colunas ==
   if ( nY1 != nY2 )
   	fIncX12 = INCREMENTOX ( nX1, nY1, nX2, nY2 );
   else
      bIncValido12 = ! bIncValido12;
   if ( nY1 != nY4 )
   	fIncX14 = INCREMENTOX ( nX1, nY1, nX4, nY4 );
   else
      bIncValido14 = ! bIncValido14;
   if ( nY2 != nY3 )
   	fIncX23 = INCREMENTOX ( nX2, nY2, nX3, nY3 );
   else
      bIncValido23 = ! bIncValido23;
   if ( nY3 != nY4 )
   	fIncX34 = INCREMENTOX ( nX3, nY3, nX4, nY4 );
   else
      bIncValido34 = ! bIncValido34;

   // varre poligono do extremo superior ateh o extremo inferior
   INT16
   	nXIni,    // deve possuir apenas x inicial da linha horizontal
      nXFin,    // deve possuir apenas x final da linha horizontal
      nX;       // valor x da interceccao entre duas retas
	for ( INT16 ncY = nYSup; ncY < nYInf; ncY ++ ) {

      // valores extremos iniciais de x
   	nXIni = MAXX + 1;
   	nXFin = MINX - 1;

      // se estao sendo definidas posicoes de x entre vertices 1 e 4
      if ( bIncValido14 &&
		     ( ncY >= nY1 && ncY <= nY4 ) ||
		     ( ncY <= nY1 && ncY >= nY4 ) ) {

			nX = (INT16) ( (float) ( ncY - nY4 ) * fIncX14 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 1 e 2
      if ( bIncValido12 &&
		     ( ncY >= nY1 && ncY <= nY2 ) ||
		     ( ncY <= nY1 && ncY >= nY2 ) ) {

			nX = (INT16) ( (float) ( ncY - nY2 ) * fIncX12 ) + nX2;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 2 e 3
      if ( bIncValido23 &&
		     ( ncY >= nY2 && ncY <= nY3 ) ||
		     ( ncY <= nY2 && ncY >= nY3 ) ){

			nX = (INT16) ( (float) ( ncY - nY3 ) * fIncX23 ) + nX3;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 3 e 4
      if ( bIncValido34 &&
		     ( ncY >= nY3 && ncY <= nY4 ) ||
		     ( ncY <= nY3 && ncY >= nY4 ) ) {

			nX = (INT16) ( (float) ( ncY - nY4 ) * fIncX34 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

		// verifica se eh para trocar x inicial e final, se sim troca
		if ( nXIni > nXFin ) {

			INT16
				nX = nXIni;
			nXIni = nXFin;
			nXFin = nX;
		}

      // faz reta horizontal se foram definidos valores acima
		if ( nXIni <= nXFin )
			memset ( (void*) ( m_pyVr [ m_nVrAtiva ] + RESX * ncY + nXIni ), yCor, nXFin - nXIni );
      else {
      	printf ( MSG_ERROCALCX );
         getch ();
      }
	}
}



/* 2 - Faz poligono na memoria virtual de video usando
incremento de x para calcular posicao de x. Entrada eh objeto poligono.
**********************************************************/
void CGr::mAtrVrPol2D ( CPol2D &oPol2D, BYTE yCor )
{
	mAtrVrPol2D ( oPol2D.m_anX [ 0 ], oPol2D.m_anY [ 0 ],
                  oPol2D.m_anX [ 1 ], oPol2D.m_anY [ 1 ],
                  oPol2D.m_anX [ 2 ], oPol2D.m_anY [ 2 ],
                  oPol2D.m_anX [ 3 ], oPol2D.m_anY [ 3 ],
                  yCor );
}



/* 1 - Desenha um poligono na memoria virtual de video usando
a declividade para calcular interseccoes.
Entrada eh objeto 2D. Para garantir desenho de poligino convexo
os vertices adjacentes devem ocupar posicoes contiguas nas matrizes
de vertices do objeto poligono. Ou seja, visualmente devem ser
definidos em ordem horaria ou anti-horaria.

								(x4, y4)
								*                -
					        / \                 loop
							 /   \                    |
				(x1, y1) *     \                   |
							 \		 * (x3, y3)         | vai calculando interseccoes
					        \	/                   | entre retas horizontais e
								\ /                    V retas das bordas do poligono
								 *               -
								 (x2, y2)

     Formula das retas (usada para definir a e b):
	        Y = aX + b, a eh declividade
                       b eh deslocamento da reta no eixo y
     Formula usada nas interseccoes:
	        X = ( Y - b ) / a

Os vertices sao numerados de 1 a 4 correspondendo a sua ordem
na matriz de vertices do objeto poligono.
Sao chamados vertices superior ou inferior os vertices dos extremos
considerando y.
**********************************************************/
void CGr::mAtrVrPol2DD ( INT nX1, INT nY1, INT nX2, INT nY2, INT nX3, INT nY3,
								  INT nX4, INT nY4, BYTE yCor )
{
   // recorta poligono, faz com que vertices permanecam dentro da tela
	// verifica se os X estao fora dos limites, se estao os corrige
   // testa x1
	if ( nX1 < MINX )
		nX1 = MINX;
	else
		if ( nX1 > MAXX )
			nX1 = MAXX;
   // testa x2
	if ( nX2 < MINX )
		nX2 = MINX;
	else
		if ( nX2 > MAXX )
			nX2 = MAXX;
   // testa x3
	if ( nX3 < MINX )
		nX3 = MINX;
	else
		if ( nX3 > MAXX )
			nX3 = MAXX;
   // testa x4
	if ( nX4 < MINX )
		nX4 = MINX;
	else
		if ( nX4 > MAXX )
			nX4 = MAXX;
	// verifica se os Y estao fora dos limites, se estao os corrige
   // testa y1
	if ( nY1 < MINY )
		nY1 = MINY;
	else
		if ( nY1 > MAXY )
			nY1 = MAXY;
   // testa y2
	if ( nY2 < MINY )
		nY2 = MINY;
	else
		if ( nY2 > MAXY )
			nY2 = MAXY;
   // testa y3
	if ( nY3 < MINY )
		nY3 = MINY;
	else
		if ( nY3 > MAXY )
			nY3 = MAXY;
   // testa y4
	if ( nY4 < MINY )
		nY4 = MINY;
	else
		if ( nY4 > MAXY )
			nY4 = MAXY;

	// define maior e menor Y (vertices extremos em y).
	INT
		nYSup = nY1,
		nYInf = nY2;
   // se o valor do y superior for mais abaixo, na tela, entao muda
   if ( nYSup > nY1 )
      nYSup = nY1;
   if ( nYSup > nY2 )
      nYSup = nY2;
   if ( nYSup > nY3 )
      nYSup = nY3;
   if ( nYSup > nY4 )
      nYSup = nY4;
   // se o valor do y inferior for mais acima, na tela, entao muda
   if ( nYInf < nY1 )
      nYInf = nY1;
   if ( nYInf < nY2 )
      nYInf = nY2;
   if ( nYInf < nY3 )
      nYInf = nY3;
   if ( nYInf < nY4 )
      nYInf = nY4;

   // define declividades e altura em que as retas cortam do eixo dos y
   float
   	fDec01 = 0,
   	fDec03 = 0,
   	fDec12 = 0,
   	fDec23 = 0,
      nB01,
      nB03,
      nB12,
      nB23;
   // somente sao calculados valores em que a declividade nao seja infinito
   if ( nX1 != nX2 ) {
   	fDec01 = DECLIVIDADE ( nX1, nY1, nX2, nY2 );
      nB01   = nY1 - (INT16) ( fDec01 * (float) nX1 );
   }
   if ( nX1 != nX4 ) {
   	fDec03 = DECLIVIDADE ( nX1, nY1, nX4, nY4 );
      nB03   = nY1 - (INT16) ( fDec03 * (float) nX1 );
   }
   if ( nX2 != nX3 ) {
   	fDec12 = DECLIVIDADE ( nX2, nY2, nX3, nY3 );
      nB12   = nY2 - (INT16) ( fDec12 * (float) nX2 );
   }
   if ( nX3 != nX4 ) {
   	fDec23 = DECLIVIDADE ( nX3, nY3, nX4, nY4 );
      nB23   = nY3 - (INT16) ( fDec23 * (float) nX3 );
   }

   // varre poligono do extremo superior ateh o extremo inferior
   INT16
   	nXIni,    // deve possuir apenas x inicial da linha horizontal
      nXFin,    // deve possuir apenas x final da linha horizontal
      nX;       // valor x da interceccao entre duas retas
	for ( INT16 ncY = nYSup; ncY < nYInf; ncY ++ ) {

      // valores extremos iniciais de x
   	nXIni = MAXX + 1;
   	nXFin = MINX - 1;

      // se estah sendo definida interseccoes entre vertice 0 e 3
      if ( ncY >= nY1 && ncY <= nY4 ||
		     ncY <= nY1 && ncY >= nY4 ) {

         // interseccao reta 0 a 3
         if ( fDec03 == 0 ) {
            // verifica se eh para alterar x inicial
			   if ( nX1 < nXIni )
				   nXIni = nX1;
	         // verifica se eh para alterar x final
			   if ( nX1 > nXFin )
				   nXFin = nX1;
            // verifica se eh para alterar x inicial
			   if ( nX4 < nXIni )
				   nXIni = nX4;
	         // verifica se eh para alterar x final
			   if ( nX4 > nXFin )
				   nXFin = nX4;
         }
         else {
	      	nX = (INT16) ( ( ncY - nB03 ) / fDec03 );

            // verifica se nao houve erro de arredondamento, se houve corrige
            if ( nX < nX1 && nX < nX4 )
            	if ( nX1 < nX4 )
               	nX = nX1;
               else
               	nX = nX4;

            // se o novo valor calculado de x define uma interseccao visivel
            // verifica se eh para alterar x inicial
			   if ( nX < nXIni )
				   nXIni = nX;
	         // verifica se eh para alterar x final
			   if ( nX > nXFin )
				   nXFin = nX;
         }
      }

      // se estah sendo definida interseccoes entre vertice 0 e 1
      if ( ncY >= nY1 && ncY <= nY2 ||
		     ncY <= nY1 && ncY >= nY2 ) {

         // interseccao reta 0 a 1
         if ( fDec01 == 0 ) {
            // verifica se eh para alterar x inicial
			   if ( nX1 < nXIni )
				   nXIni = nX1;
	         // verifica se eh para alterar x final
			   if ( nX1 > nXFin )
				   nXFin = nX1;
            // verifica se eh para alterar x inicial
			   if ( nX2 < nXIni )
				   nXIni = nX2;
	         // verifica se eh para alterar x final
			   if ( nX2 > nXFin )
				   nXFin = nX2;
         }
         else {
	      	nX = (INT16) ( ( ncY - nB01 ) / fDec01 );

            // verifica se nao houve erro de arredondamento, se houve corrige
            if ( nX < nX1 && nX < nX2 )
            	if ( nX1 < nX2 )
               	nX = nX1;
               else
               	nX = nX2;

            // se o novo valor calculado de x define uma interseccao visivel
            // verifica se eh para alterar x inicial
			   if ( nX < nXIni )
				   nXIni = nX;
	         // verifica se eh para alterar x final
			   if ( nX > nXFin )
				   nXFin = nX;
         }
      }

      // se estah sendo definida interseccoes entre vertice 1 e 2
      if ( ncY >= nY2 && ncY <= nY3  ||
		     ncY <= nY2 && ncY >= nY3 ) {

         // interseccao reta 1 a 2
         if ( fDec12 == 0 ) {
            // verifica se eh para alterar x inicial
			   if ( nX2 < nXIni )
				   nXIni = nX2;
	         // verifica se eh para alterar x final
			   if ( nX2 > nXFin )
				   nXFin = nX2;
            // verifica se eh para alterar x inicial
			   if ( nX3 < nXIni )
				   nXIni = nX3;
	         // verifica se eh para alterar x final
			   if ( nX3 > nXFin )
				   nXFin = nX3;
         }
         else {
	      	nX = (INT16) ( ( ncY - nB12 ) / fDec12 );

            // verifica se nao houve erro de arredondamento, se houve corrige
            if ( nX < nX2 && nX < nX3 )
            	if ( nX2 < nX3 )
               	nX = nX2;
               else
               	nX = nX3;

            // se o novo valor calculado de x define uma interseccao visivel
            // verifica se eh para alterar x inicial
			   if ( nX < nXIni )
				   nXIni = nX;
	         // verifica se eh para alterar x final
			   if ( nX > nXFin )
				   nXFin = nX;
         }
      }

      // se estah sendo definida interseccoes entre vertice 2 e 3
      if ( ncY >= nY3 && ncY <= nY4 ||
		     ncY <= nY3 && ncY >= nY4 ) {

         // interseccao reta 2 a 3
         if ( fDec23 == 0 ) {
            // verifica se eh para alterar x inicial
			   if ( nX3 < nXIni )
				   nXIni = nX3;
	         // verifica se eh para alterar x final
			   if ( nX3 > nXFin )
				   nXFin = nX3;
            // verifica se eh para alterar x inicial
			   if ( nX4 < nXIni )
				   nXIni = nX4;
	         // verifica se eh para alterar x final
			   if ( nX4 > nXFin )
				   nXFin = nX4;
         }
         else {
	      	nX = (INT16) ( ( ncY - nB23 ) / fDec23 );

            // verifica se nao houve erro de arredondamento, se houve corrige
            if ( nX < nX3 && nX < nX4 )
            	if ( nX3 < nX4 )
               	nX = nX3;
               else
               	nX = nX4;

            // se o novo valor calculado de x define uma interseccao visivel
            // verifica se eh para alterar x inicial
			   if ( nX < nXIni )
				   nXIni = nX;
	         // verifica se eh para alterar x final
			   if ( nX > nXFin )
				   nXFin = nX;
         }
      }

		// verifica se eh para trocar x inicial e final, se sim troca
		if ( nXIni > nXFin ) {

			INT16
				nX = nXIni;
			nXIni = nXFin;
			nXFin = nX;
		}

      // recorta x: verifica se os X estao fora dos limites
		if ( nXIni < MINX || nXIni > MAXX ) {
			nXIni = MINX;
         printf ( MSG_RECORTEXMIN );
         getch ();
      }
		if ( nXFin < MINX || nXFin > MAXX ) {
			nXFin = MAXX;
         printf ( MSG_RECORTEXMAX );
         getch ();
      }

      // faz reta horizontal se foram definidos valores acima
		if ( nXIni <= nXFin )
			memset ( (void*) ( m_pyVr [ m_nVrAtiva ] + RESX * ncY + nXIni ), yCor, nXFin - nXIni );
	}
}



/* 2 - Desenha um poligono na memoria virtual de video usando
a declividade para calcular interseccoes. Entrada eh vertice a vertice.
**********************************************************/
void CGr::mAtrVrPol2DD ( CPol2D &oPol2D, BYTE yCor )
{
	mAtrVrPol2DD ( oPol2D.m_anX [ 0 ], oPol2D.m_anY [ 0 ],
                   oPol2D.m_anX [ 1 ], oPol2D.m_anY [ 1 ],
                   oPol2D.m_anX [ 2 ], oPol2D.m_anY [ 2 ],
                   oPol2D.m_anX [ 3 ], oPol2D.m_anY [ 3 ],
                   yCor );
}
